-- moved to sum-pyramid project
module Numeric.LAPACK.Example.SumPyramid where

import qualified Numeric.LAPACK.Matrix.Triangular as Triangular
-- import qualified Numeric.LAPACK.Matrix.Square as Square
import qualified Numeric.LAPACK.Matrix.Layout as Layout
import qualified Numeric.LAPACK.Matrix as Matrix
import qualified Numeric.LAPACK.Singular as Singular
import qualified Numeric.LAPACK.Vector as Vector
import Numeric.LAPACK.Matrix (ShapeInt, (#*|))
-- import Numeric.LAPACK.Matrix (ShapeInt, (#+#), (#-#), (#\|), (#*|))
-- import Numeric.LAPACK.Vector (Vector, (|+|), (|-|))
import Numeric.LAPACK.Format ((##))

import qualified Combinatorics

import qualified Data.Array.Comfort.Boxed as BoxedArray
import qualified Data.Array.Comfort.Storable as Array
import qualified Data.Array.Comfort.Shape as Shape
import qualified Data.IntSet as IntSet
import qualified Data.Set as Set
import Data.Array.Comfort.Storable (Array)
-- import Foreign.Storable (Storable)

import Data.Foldable (for_)
import Data.Tuple.HT (mapSnd)

{-
import qualified Control.Monad.Trans.State as MS

import qualified Test.QuickCheck as QC
-}



example0 :: [Double]
example0 = [36,42,16,28,15]

{-
        ...
      ... ...
    ... ... ...
  ... 120 ... 120
.28 .72 ... .22 ...
-}
example2 :: [[Maybe Double]]
example2 =
   let __ = Nothing; d = Just in
                [__] :
             [__,    __] :
          [__,   __,    __] :
       [__,  d 120,  __,  d 120] :
   [d 28, d 72,  __,   d 22,  __] :
   []

pyramid ::
   Array ShapeInt Double ->
   Array (Shape.LowerTriangular ShapeInt) Double
pyramid xs =
   let shape = Array.shape xs
--       baseRow = last $ Shape.indices shape
       baseRow = Shape.size shape - 1
       arr =
         fmap (\(i,j) ->
            if i==baseRow
               then xs Array.! j
               else arr BoxedArray.! (i+1,j) + arr BoxedArray.! (i+1,j+1)) $
         BoxedArray.indices $ Shape.lowerTriangular shape
   in Array.fromBoxed arr
--   in BoxedArray.toStorable arr

basis ::
   ShapeInt -> Matrix.General ShapeInt (Shape.LowerTriangular ShapeInt) Double
basis shape@(Shape.ZeroBased n) =
   Matrix.fromRows (Shape.lowerTriangular shape) $
   map (pyramid . Vector.unit shape) $ take n [0..]

solve ::
   [[Maybe Double]] ->
   (Int, Array (Shape.LowerTriangular ShapeInt) Double)
solve puzzle =
   let n = length puzzle
       indexed = do
         (i, xs) <- zip [0..] puzzle
         (j, Just x) <- zip [0..] xs
         return ((i,j),x)
       fullBasis = Matrix.transpose $ basis (Matrix.shapeInt n)
       selected =
         Matrix.takeRowArray
            (BoxedArray.vectorFromList (map fst indexed))
            fullBasis
   in mapSnd ((fullBasis #*|) . Matrix.flattenColumn)
         (Singular.leastSquaresMinimumNormRCond 1e-5 selected $
          Matrix.singleColumn Layout.ColumnMajor $
          Vector.autoFromList (map snd indexed))

solvable :: Int -> [(Int,Int)] -> Bool
solvable n =
   let shape = Matrix.shapeInt n
       fullBasis = basis shape
   in \ixs ->
         (n==) $ length $ takeWhile (1e-5<) $ Vector.toList $
         (#*| Vector.one shape) $ Singular.values $
         Matrix.takeColumnArray (BoxedArray.vectorFromList ixs) fullBasis

solvables :: Int -> [[(Int,Int)]]
solvables n =
   let check = solvable n
   in filter check $
      Combinatorics.tuples n $ Shape.indices $
      Shape.lowerTriangular $ Matrix.shapeInt n

{-
Check, whether a sum pyramid contains a sub-pyramid of size k
with more than k given fields.
If yes, then the pyramid has redundancies in a sub-pyramid
and is not solvable.
-}
wellcrowded :: Int -> [(Int,Int)] -> Bool
wellcrowded n ixs =
   let triShape = Shape.lowerTriangular $ Matrix.shapeInt n
       set = Set.fromList ixs
       countSubTriangle (i,j) k =
         length $
         filter (\(si,sj) -> Set.member (i+si,j+sj) set) $
         Shape.indices $ Shape.lowerTriangular $ Matrix.shapeInt k
   in and $ do
         (i,j) <- Shape.indices triShape
         k <- [0..n-i]
         return $ countSubTriangle (i,j) k <= k

wellcrowdedIntSet :: Int -> [(Int,Int)] -> Bool
wellcrowdedIntSet n ixs =
   let triShape = Shape.lowerTriangular $ Matrix.shapeInt n
       set = IntSet.fromList $ map (Shape.offset triShape) ixs
       countSubTriangle (i,j) k =
         length $
         filter (\(si,sj) ->
            flip IntSet.member set $ Shape.offset triShape (i+si,j+sj)) $
         Shape.indices $ Shape.lowerTriangular $ Matrix.shapeInt k
   in and $ do
         (i,j) <- Shape.indices triShape
         k <- [0..n-i]
         return $ countSubTriangle (i,j) k <= k


{-
Check whether the linear independence criterion matches
the subpyramid criterion 'wellcrowded'.
Well, it does not, the smallest counterexample is:

   *
  . .
 . * .
* . . *
-}
counterexamples :: Int -> [[(Int, Int)]]
counterexamples n =
   let check = solvable n
   in filter (\ixs -> check ixs /= wellcrowded n ixs) $
      Combinatorics.tuples n $ Shape.indices $
      Shape.lowerTriangular $ Matrix.shapeInt n


boolMatrix :: Int -> [(Int, Int)] -> Matrix.Lower ShapeInt Float
boolMatrix n ixs =
   Triangular.fromLowerRowMajor $
   Array.fromAssociations 0
      (Shape.lowerTriangular $ Matrix.shapeInt n)
      (map (\ix -> (ix,1::Float)) ixs)


main :: IO ()
main = do
   Triangular.fromLowerRowMajor (pyramid (Vector.autoFromList example0))
      ## "%.0f"

   mapSnd Triangular.fromLowerRowMajor (solve example2) ## "%.0f"

   putStrLn "\nsolvable:"
   let n = 3
    in for_ (solvables n) $ \ixs -> do
          putStrLn ""
          boolMatrix n ixs ## "%.0f"

   putStrLn "\nunsolvable:"
   for_ [0..5] $ \n ->
      for_ (counterexamples n) $ \ixs -> do
         putStrLn ""
         boolMatrix n ixs ## "%.0f"

   -- 1,1,3,17,149,1824,29001,573549,13604001
   mapM_ (print . length . solvables) [0..]

{-
https://oeis.org/A014068

map (\n -> Comb.binomial (div (n*(n+1)) 2) n) [1..10::Integer]

Possibilities of choosing n numbers from a n-sized pyramid.
-}
